import os
import sys
from contextlib import contextmanager


@contextmanager
def stdout_redirected(to=os.devnull):
    """
    A context manager to temporarily redirect stdout to another file. This is useful when you want to suppress the
    output from blender, especially when you are rendering a large number of images.

    Example usage:
    >>> import os
    >>> filename = os.devnull
    >>> with stdout_redirected(to=filename):
    ...     print("from Python")
    ...     os.system("echo non-Python applications are also supported")

    :param to: The file to redirect stdout to. Default is os.devnull.
    :return: The context manager.
    """

    fd = sys.stdout.fileno()

    def _redirect_stdout(_to):
        sys.stdout.close()  # + implicit flush()
        os.dup2(_to.fileno(), fd)  # fd writes to '_to' file
        sys.stdout = os.fdopen(fd, 'w')  # Python writes to fd

    with os.fdopen(os.dup(fd), 'w') as old_stdout:
        with open(to, 'w') as file:
            _redirect_stdout(_to=file)
        try:
            yield  # allow code to be run with the redirected stdout
        finally:
            _redirect_stdout(_to=old_stdout)  # restore stdout.



from typing import Literal
import bpy
from bpy_helper.scene import reset_scene, import_3d_model, normalize_scene, scene_meshes


def uv_unwrap(
    input_mesh_file: str,
    output_mesh_file: str,
    method: Literal['unwrap', 'smart_project', 'cube_project', 'cylinder_project', 'sphere_project'] = 'unwrap',
    normalize: bool = True,
    **kwargs
) -> bool:
    """
    Unwraps the UVs of the given mesh file.

    Args:
        input_mesh_file (str): The path to the input mesh file.
        output_mesh_file (str): The path to the output mesh file.
        method (Literal['unwrap', 'smart_project', 'cube_project', 'cylinder_project', 'sphere_project'], optional): The method to use for unwrapping. Defaults to 'unwrap'.
        normalize (bool, optional): Whether to normalize the scene before unwrapping. Defaults to True.
        kwargs: Additional keyword arguments.

    Returns:
        bool: True if the unwrapping was successful, False otherwise.
    """

    # clean the scene
    reset_scene()

    import_3d_model(input_mesh_file)
    # turn to Z-up Y-forward
    for obj in scene_meshes():
        obj.rotation_mode = 'XYZ'
        obj.rotation_euler = (0.0, 0.0, 0.0)

    # normalize the scene
    if normalize:
        scale, offset = normalize_scene(use_bounding_sphere=True)
        print(f"scale: {scale}, offset: {offset}")

    bpy.ops.object.select_all(action='DESELECT')
    bpy.ops.object.mode_set(mode='EDIT')
    bpy.ops.mesh.select_all(action='SELECT')

    if method == 'unwrap':
        _kwargs = {
            'method': 'ANGLE_BASED',
            'margin': 0.001
        }
        _kwargs.update(kwargs)
        bpy.ops.uv.unwrap(**_kwargs)
    elif method == 'smart_project':
        _kwargs = {
        }
        _kwargs.update(kwargs)
        bpy.ops.uv.smart_project(**_kwargs)
    elif method == 'cube_project':
        _kwargs = {
        }
        _kwargs.update(kwargs)
        bpy.ops.uv.cube_project(**_kwargs)
    elif method == 'cylinder_project':
        _kwargs = {
        }
        _kwargs.update(kwargs)
        bpy.ops.uv.cylinder_project(**_kwargs)
    elif method == 'sphere_project':
        _kwargs = {
        }
        _kwargs.update(kwargs)
        bpy.ops.uv.sphere_project(**_kwargs)
    else:
        raise ValueError(f"Unsupported method: {method}")

    bpy.ops.object.mode_set(mode='OBJECT')

    # export to same coordinates
    bpy.ops.wm.obj_export(
        filepath=output_mesh_file,
        export_uv=True,
        export_normals=False,
        export_colors=False,
        export_materials=False,
        export_triangulated_mesh=True,
        up_axis='Z',
        forward_axis='Y'
    )

    return True



from typing import List, Optional

import bpy


def save_blend_file(filepath) -> None:
    """
    Save the current blend file

    :param filepath: file path to save
    """

    bpy.ops.wm.save_as_mainfile(filepath=filepath)


# adapted from BlenderProc
def get_nodes_created_in_func(nodes: List[bpy.types.Node], created_in_func: str) -> List[bpy.types.Node]:
    """
    Returns all nodes which are created in the given function

    :param nodes: list of nodes of the current material
    :param created_in_func: return all nodes created in the given function
    :return: The list of nodes with the given type.
    """

    return [node for node in nodes if "created_in_func" in node and node["created_in_func"] == created_in_func]


def get_nodes_with_type(nodes: List[bpy.types.Node], node_type: str,
                        created_in_func: Optional[str] = None) -> List[bpy.types.Node]:
    """
    Returns all nodes which are of the given node_type

    :param nodes: list of nodes of the current material
    :param node_type: node types
    :param created_in_func: Only return nodes created by the specified function
    :return: list of nodes, which belong to the type
    """

    nodes_with_type = [node for node in nodes if node_type in node.bl_idname]
    if created_in_func:
        nodes_with_type = get_nodes_created_in_func(nodes_with_type, created_in_func)
    return nodes_with_type


def get_the_one_node_with_type(nodes: List[bpy.types.Node], node_type: str,
                               created_in_func: str = "") -> bpy.types.Node:
    """
    Returns the one node which is of the given node_type

    This function will only work if there is only one of the nodes of this type.

    :param nodes: list of nodes of the current material
    :param node_type: node types
    :param created_in_func: only return node created by the specified function
    :return: node of the node type
    """

    node = get_nodes_with_type(nodes, node_type, created_in_func)
    if node and len(node) == 1:
        return node[0]
    raise RuntimeError(f"There is not only one node of this type: {node_type}, there are: {len(node)}")


def render_depth_map(output_dir, file_prefix='depth') -> None:
    """
    Render depth map

    :param output_dir: output directory
    :param file_prefix: file prefix, default is 'depth'
    """

    # disable material override
    bpy.context.scene.view_layers["ViewLayer"].material_override = None

    bpy.context.scene.render.use_compositing = True
    bpy.context.scene.use_nodes = True

    tree = bpy.context.scene.node_tree
    links = tree.links
    # Use existing render layer
    render_layer_node = get_the_one_node_with_type(tree.nodes, 'CompositorNodeRLayers')

    # Enable z-buffer pass
    bpy.context.view_layer.use_pass_z = True

    # Build output node
    output_file = tree.nodes.new("CompositorNodeOutputFile")
    output_file.base_path = output_dir
    output_file.format.file_format = "OPEN_EXR"
    # set a different path (in case overwrite last file)
    bpy.context.scene.render.filepath = f'{output_dir}/rgb_for_{file_prefix}.png'
    output_file.file_slots.values()[0].path = file_prefix

    # Feed the Z-Buffer output of the render layer to the input of the file IO layer
    links.new(render_layer_node.outputs["Depth"], output_file.inputs['Image'])
    bpy.ops.render.render(animation=False, write_still=True)

    # Clean up
    for link in output_file.inputs[0].links:
        links.remove(link)
    tree.nodes.remove(output_file)
    bpy.context.scene.render.use_compositing = False
    bpy.context.scene.use_nodes = False
    bpy.context.view_layer.use_pass_z = False


def render_normal_map(output_dir, file_prefix="normal") -> None:
    """
    Render normal map

    :param output_dir: output directory
    :param file_prefix: file prefix, default is 'normal'
    """

    # disable material override
    bpy.context.scene.view_layers["ViewLayer"].material_override = None

    bpy.context.scene.render.use_compositing = True
    bpy.context.scene.use_nodes = True

    tree = bpy.context.scene.node_tree
    links = tree.links
    # Use existing render layer
    render_layer_node = get_the_one_node_with_type(tree.nodes, 'CompositorNodeRLayers')

    # Enable normal pass
    bpy.context.view_layer.use_pass_normal = True

    # Separate into RGB
    separate_rgba = tree.nodes.new("CompositorNodeSepRGBA")
    links.new(render_layer_node.outputs["Normal"], separate_rgba.inputs["Image"])

    combine_rgba = tree.nodes.new("CompositorNodeCombRGBA")
    for row_index in range(3):
        map_range = tree.nodes.new("CompositorNodeMapRange")
        map_range.inputs["From Min"].default_value = -1.0
        map_range.inputs["From Max"].default_value = 1.0
        map_range.inputs["To Min"].default_value = 0.0
        map_range.inputs["To Max"].default_value = 1.0
        links.new(separate_rgba.outputs[row_index], map_range.inputs["Value"])
        links.new(map_range.outputs["Value"], combine_rgba.inputs[row_index])

    # Build output node
    output_file = tree.nodes.new("CompositorNodeOutputFile")
    output_file.base_path = output_dir
    output_file.format.file_format = "OPEN_EXR"
    # set a different path (in case overwrite last file)
    bpy.context.scene.render.filepath = f'{output_dir}/rgb_for_{file_prefix}.png'
    output_file.file_slots.values()[0].path = file_prefix

    # Feed the combined rgb output to the input of the file IO layer
    links.new(combine_rgba.outputs["Image"], output_file.inputs['Image'])
    bpy.ops.render.render(animation=False, write_still=True)

    # Clean up
    for link in output_file.inputs[0].links:
        links.remove(link)
    tree.nodes.remove(output_file)
    bpy.context.scene.render.use_compositing = False
    bpy.context.scene.use_nodes = False
    bpy.context.view_layer.use_pass_normal = False


# some helper functions
mat2list = lambda x: [[float(xxx) for xxx in xx] for xx in x]
array2list = lambda x: [float(xx) for xx in x]



import math

import bpy
import mathutils


def remove_default_objects() -> None:
    """
    Removes the default objects from the scene.
    """

    objects_to_remove = ["Cube", "Camera", "Light"]

    for obj_name in objects_to_remove:
        bpy.ops.object.select_all(action='DESELECT')
        obj = bpy.data.objects.get(obj_name)
        if obj is not None:
            obj.select_set(True)
            bpy.ops.object.delete()


def reset_scene() -> None:
    """
    Resets the scene to the default state.
    """

    bpy.ops.wm.read_homefile()

    # Remove the default cube
    remove_default_objects()


def import_3d_model(object_path) -> None:
    """
    Loads a 3d model into the scene.
    """

    if object_path.endswith(".glb"):
        bpy.ops.import_scene.gltf(filepath=object_path, merge_vertices=True)
    elif object_path.endswith(".obj"):
        bpy.ops.wm.obj_import(filepath=object_path)
    elif object_path.endswith(".fbx"):
        bpy.ops.import_scene.fbx(filepath=object_path)
    elif object_path.endswith(".blend"):
        bpy.ops.wm.open_mainfile(filepath=object_path)
    elif object_path.endswith(".ply"):
        bpy.ops.import_mesh.ply(filepath=object_path)
    else:
        raise ValueError(f"Unsupported file type: {object_path}")


def scene_root_objects():
    """
    Yields all root objects in the scene.
    """

    for obj in bpy.context.scene.objects.values():
        if not obj.parent:
            yield obj


def scene_meshes():
    """
    Yields all mesh objects in the scene.
    """

    for obj in bpy.context.scene.objects.values():
        if isinstance(obj.data, bpy.types.Mesh):
            yield obj


def scene_bbox(single_obj=None, ignore_matrix=False) -> tuple[mathutils.Vector, mathutils.Vector]:
    """
    Compute the bounding box of the scene.

    :param single_obj: if not None, only compute the bounding box for this object
    :param ignore_matrix: if True, ignore the transformation matrix of the object
    :return: bounding box min and max
    """

    bbox_min = (math.inf,) * 3
    bbox_max = (-math.inf,) * 3
    found = False
    for obj in scene_meshes() if single_obj is None else [single_obj]:
        found = True
        for coord in obj.bound_box:
            coord = mathutils.Vector(coord)
            if not ignore_matrix:
                coord = obj.matrix_world @ coord
            bbox_min = tuple(min(x, y) for x, y in zip(bbox_min, coord))
            bbox_max = tuple(max(x, y) for x, y in zip(bbox_max, coord))
    if not found:
        raise RuntimeError("no objects in scene to compute bounding box for")
    return mathutils.Vector(bbox_min), mathutils.Vector(bbox_max)


def get_center(l) -> float:
    """
    Compute the center value of a list of numbers.
    """

    return (max(l) + min(l)) / 2 if l else 0.0


def scene_sphere(single_obj=None, ignore_matrix=False) -> tuple[mathutils.Vector, float]:
    """
    Compute the bounding sphere of the scene.

    :param single_obj: if not None, only compute the bounding sphere for this object
    :param ignore_matrix: if True, ignore the transformation matrix of the object
    :return: bounding sphere center and radius
    """

    found = False
    points_co_global = []
    for obj in scene_meshes() if single_obj is None else [single_obj]:
        found = True
        mesh = obj.data
        for vertex in mesh.vertices:
            vertex_co = vertex.co
            if not ignore_matrix:
                vertex_co = obj.matrix_world @ vertex_co
            points_co_global.extend([vertex_co])
    if not found:
        raise RuntimeError("no objects in scene to compute bounding sphere for")
    x, y, z = [[point_co[i] for point_co in points_co_global] for i in range(3)]
    b_sphere_center = mathutils.Vector([get_center(axis) for axis in [x, y, z]]) if (x and y and z) else None
    b_sphere_radius = max(((point - b_sphere_center) for point in points_co_global)) if b_sphere_center else None
    return b_sphere_center, b_sphere_radius.length


def normalize_scene(scale=None, offset=None, use_bounding_sphere=False, target_scale=0.5) -> tuple[float, mathutils.Vector]:
    """
    Normalize the scene by scaling and translating all objects.

    :param scale: scale factor
    :param offset: translation offset
    :param use_bounding_sphere: if True, use the bounding sphere to compute the scale factor
    :param target_scale: the target scale factor
    :return: scale factor and translation offset
    """

    if scale is None:
        if not use_bounding_sphere:
            bbox_min, bbox_max = scene_bbox()
            scale = target_scale / max(bbox_max - bbox_min)
        else:
            center, radius = scene_sphere()
            scale = target_scale / radius
    for obj in scene_root_objects():
        obj.scale = obj.scale * scale

    # Apply scale to matrix_world.
    bpy.context.view_layer.update()
    if offset is None:
        bbox_min, bbox_max = scene_bbox()
        offset = -(bbox_min + bbox_max) / 2
    for obj in scene_root_objects():
        obj.matrix_world.translation += offset
    bpy.ops.object.select_all(action="DESELECT")
    return scale, offset


def scale_from_trimesh(filename, use_bounding_sphere=False) -> float:
    """
    Compute the scale factor from a trimesh file. Some GLB models contains animation data, which can effect the scale
    factor. This function uses trimesh to compute the scale factor, which can avoid the animation data.

    :param filename: the file name of the 3D object
    :param use_bounding_sphere: if True, use the bounding sphere to compute the scale factor
    """

    import trimesh
    mesh = trimesh.load(filename, force="mesh")
    if use_bounding_sphere:
        radius = mesh.bounding_sphere.primitive.radius
        scale = 0.5 / radius
    else:
        bbox_min = mesh.bounds[0]
        bbox_max = mesh.bounds[1]
        scale = 1.0 / max(bbox_max - bbox_min)
    return scale



import os
from typing import Union

import bpy
import numpy as np
from mathutils import Euler, Vector

from bpy_helper.io import get_the_one_node_with_type


def remove_all_lights() -> None:
    """
    Remove all lights in the scene
    """

    # remove all lights
    light_objs = []
    for obj in bpy.data.objects:
        if obj.type == 'LIGHT':
            # bpy.data.objects.remove(obj)
            light_objs.append(obj)
    for obj in light_objs:
        bpy.data.objects.remove(obj)

    world = bpy.context.scene.world
    nodes = world.node_tree.nodes

    # clear the world background links
    world.use_nodes = True
    if len(world.node_tree.nodes['Background'].inputs[0].links) > 0:
        world.node_tree.links.remove(world.node_tree.nodes['Background'].inputs[0].links[0])
    # clear the world background nodes
    for node in nodes:
        if node.type in ['TEX_ENVIRONMENT', 'TEX_COORD', 'MAPPING']:
            nodes.remove(node)


def create_point_light(location, power=800., rgb=(1., 1., 1.), hard_shadow=False, keep_other_lights=False) \
        -> bpy.types.Object:
    """
    Creates a point light at the given location.

    :param location: The location of the light
    :param power: The power of the light, default is 800
    :param rgb: The color of the light, default is white
    :param hard_shadow: If true, the shadow will be hard. Default is False
    :param keep_other_lights: If true, the other lights will not be removed. Default is False
    :return: The light object
    """

    if not keep_other_lights:
        remove_all_lights()

    # Set the background ambient light
    bpy.context.scene.world.node_tree.nodes["Background"].inputs[1].default_value = 1.0

    light_data = bpy.data.lights.new(name="PointLight", type="POINT")
    light_data.energy = power
    light_data.color = rgb
    if hard_shadow:
        light_data.shadow_soft_size = 0
    light_obj = bpy.data.objects.new("PointLight", light_data)
    bpy.context.collection.objects.link(light_obj)
    light_obj.location = location
    return light_obj


def create_area_light(location, power=800., size=5., keep_other_lights=False) -> bpy.types.Object:
    """
    Creates an area light at the given location.

    :param location: The location of the light
    :param power: The power of the light, default is 800
    :param size: The size of the light, default is 5
    :param keep_other_lights: If true, the other lights will not be removed. Default is False
    :return: The light object
    """

    if not keep_other_lights:
        remove_all_lights()

    light_data = bpy.data.lights.new(name="AreaLight", type="AREA")
    light_data.size = size
    light_data.energy = power
    light_data.color = (1, 1, 1)
    light_data.shape = 'DISK'

    light_obj = bpy.data.objects.new("AreaLight", light_data)
    bpy.context.collection.objects.link(light_obj)
    light_obj.location = location

    direction = Vector([0, 0, 0]) - Vector(location)
    light_obj.rotation_euler = direction.to_track_quat('-Z', 'Y').to_euler()

    return light_obj


def create_directional_light(location, power=800., rgb=(1., 1., 1.), keep_other_lights=False) -> bpy.types.Object:
    """
    Creates a directional light at the given location.

    :param location: The location of the light
    :param power: The power of the light, default is 800
    :param rgb: The color of the light, default is white
    :param keep_other_lights: If true, the other lights will not be removed. Default is False
    :return: The light object
    """

    if not keep_other_lights:
        remove_all_lights()

    light_data = bpy.data.lights.new(name="DirectionalLight", type="SUN")
    light_data.energy = power
    light_data.color = rgb

    light_obj = bpy.data.objects.new("DirectionalLight", light_data)
    bpy.context.collection.objects.link(light_obj)
    light_obj.location = location

    direction = Vector([0, 0, 0]) - Vector(location)
    light_obj.rotation_euler = direction.to_track_quat('-Z', 'Y').to_euler()

    return light_obj


def set_env_light(path_to_hdr_file: str, strength: float = 1.0,
                  rotation_euler: Union[list, Euler, np.ndarray] = None, keep_other_lights: bool = False) -> None:
    """
    Sets the world background to the given hdr_file.

    :param path_to_hdr_file: Path to the .exr file
    :param strength: The brightness of the background.
    :param rotation_euler: The euler angles of the background.
    :param keep_other_lights: If true, the other lights will not be removed.
    """

    if not keep_other_lights:
        remove_all_lights()

    scene = bpy.context.scene
    world = scene.world
    nodes = world.node_tree.nodes
    links = world.node_tree.links

    if rotation_euler is None:
        rotation_euler = [0.0, 0.0, 0.0]

    if not os.path.exists(path_to_hdr_file):
        raise FileNotFoundError(f"The given path does not exists: {path_to_hdr_file}")

    # add a texture node and load the image and link it
    texture_node = nodes.new(type="ShaderNodeTexEnvironment")
    texture_node.image = bpy.data.images.load(path_to_hdr_file, check_existing=True)

    # get the one background node of the world shader
    background_node = get_the_one_node_with_type(nodes, "Background")

    # link the new texture node to the background
    links.new(texture_node.outputs["Color"], background_node.inputs["Color"])

    # Set the brightness of the background
    background_node.inputs["Strength"].default_value = strength

    # add a mapping node and a texture coordinate node
    mapping_node = nodes.new("ShaderNodeMapping")
    tex_coords_node = nodes.new("ShaderNodeTexCoord")

    # link the texture coordinate node to mapping node
    links.new(tex_coords_node.outputs["Generated"], mapping_node.inputs["Vector"])

    # link the mapping node to the texture node
    links.new(mapping_node.outputs["Vector"], texture_node.inputs["Vector"])

    mapping_node.inputs["Rotation"].default_value = rotation_euler



import bpy
from importlib.metadata import version

IS_BLENDER_4 = version('bpy').split('.')[0] == '4'


def clear_emission_and_alpha_nodes() -> None:
    """
    Clear all emission and alpha nodes in the scene, including all objects' materials
    """

    # Loop through all materials in the scene
    for material in bpy.data.materials:
        # Check if the material has an emission shader
        if material.use_nodes:
            for node in material.node_tree.nodes:
                if 'Emission Strength' in node.inputs.keys():
                    while node.inputs['Emission Strength'].is_linked:
                        material.node_tree.links.remove(node.inputs['Emission Strength'].links[0])
                    node.inputs['Emission Strength'].default_value = 0.
                if 'Alpha' in node.inputs.keys():
                    while node.inputs['Alpha'].is_linked:
                        material.node_tree.links.remove(node.inputs['Alpha'].links[0])
                    node.inputs['Alpha'].default_value = 1.
                if node.type == 'EMISSION':
                    while node.inputs['Strength'].is_linked:
                        material.node_tree.links.remove(node.inputs['Strength'].links[0])
                    node.inputs['Strength'].default_value = 0.


def override_normal_map_op(material, normal_map_path) -> None:
    """
    Override normal map for the material (to attach a precomputed world-space normal map)
    """

    # Create texture coordinate node
    tex_coord = material.node_tree.nodes.new('ShaderNodeTexCoord')

    # Load normal exr
    normal_map = material.node_tree.nodes.new('ShaderNodeTexImage')
    normal_map.image = bpy.data.images.load(normal_map_path)
    material.node_tree.links.new(tex_coord.outputs['Window'], normal_map.inputs['Vector'])

    # Vector multiply by 2
    vector_math = material.node_tree.nodes.new('ShaderNodeVectorMath')
    vector_math.operation = 'MULTIPLY'
    vector_math.inputs[1].default_value = (2., 2., 2.)
    material.node_tree.links.new(normal_map.outputs['Color'], vector_math.inputs[0])

    # Vector subtract by 1
    vector_math_2 = material.node_tree.nodes.new('ShaderNodeVectorMath')
    vector_math_2.operation = 'SUBTRACT'
    vector_math_2.inputs[1].default_value = (1., 1., 1.)
    material.node_tree.links.new(vector_math.outputs['Vector'], vector_math_2.inputs[0])

    bsdf = material.node_tree.nodes["Principled BSDF"]
    material.node_tree.links.new(vector_math_2.outputs['Vector'], bsdf.inputs['Normal'])


def create_white_diffuse_material(override_normal_map=False, normal_map_path=None, material_name="White_Diffuse_Material") -> bpy.types.Material:
    """
    Create white diffuse material

    :param override_normal_map: override normal map
    :param normal_map_path: normal map path
    :param material_name: material name, default is "White_Diffuse_Material"
    :return: white diffuse material
    """

    material = bpy.data.materials.new(name=material_name)
    material.use_nodes = True
    bsdf = material.node_tree.nodes["Principled BSDF"]

    if IS_BLENDER_4:
        bsdf.inputs['Base Color'].default_value = (1, 1, 1, 1)
        bsdf.inputs['Specular IOR Level'].default_value = 0
        bsdf.inputs['Roughness'].default_value = 1
    else:
        bsdf.inputs['Base Color'].default_value = (1, 1, 1, 1)
        bsdf.inputs['Specular'].default_value = 0
        bsdf.inputs['Roughness'].default_value = 1

    if override_normal_map:
        override_normal_map_op(material, normal_map_path)

    return material


def create_white_emmissive_material(strength=100., override_normal_map=False, normal_map_path=None, material_name="White_Emissive_Material") -> bpy.types.Material:
    """
    Create white emissive material

    :param strength: emission strength
    :param override_normal_map: override normal map
    :param normal_map_path: normal map path
    :param material_name: material name, default is "White_Emissive_Material"
    :return: white emissive material
    """

    material = bpy.data.materials.new(name=material_name)
    material.use_nodes = True
    bsdf = material.node_tree.nodes["Principled BSDF"]

    if IS_BLENDER_4:
        bsdf.inputs['Base Color'].default_value = (1, 1, 1, 1)
        bsdf.inputs['Specular IOR Level'].default_value = 0
        bsdf.inputs['Roughness'].default_value = 1
        bsdf.inputs['Emission Color'].default_value = (1, 1, 1, 1)
        bsdf.inputs['Emission Strength'].default_value = strength
    else:
        bsdf.inputs['Base Color'].default_value = (1, 1, 1, 1)
        bsdf.inputs['Specular'].default_value = 0
        bsdf.inputs['Roughness'].default_value = 1
        bsdf.inputs['Emission'].default_value = (1, 1, 1, 1)
        bsdf.inputs['Emission Strength'].default_value = strength

    if override_normal_map:
        override_normal_map_op(material, normal_map_path)

    return material


def create_specular_ggx_material(r=0.34, override_normal_map=False, normal_map_path=None, material_name="Specular_GGX_Material") -> bpy.types.Material:
    """
    Create specular GGX material

    :param r: roughness, default is 0.34
    :param override_normal_map: whether to override normal map, default is False
    :param normal_map_path: normal map path, if override_normal_map is True
    :param material_name: material name, default is "Specular_GGX_Material"
    :return: specular GGX material
    """

    material = bpy.data.materials.new(name=material_name)
    material.use_nodes = True
    bsdf = material.node_tree.nodes["Principled BSDF"]

    if IS_BLENDER_4:
        # Keys have been remapped, ref: https://wiki.blender.org/wiki/Reference/Release_Notes/4.0/Python_API
        bsdf.inputs['Base Color'].default_value = (0, 0, 0, 1)
        bsdf.inputs['Subsurface Weight'].default_value = 0
        bsdf.inputs['Metallic'].default_value = 0
        bsdf.inputs['Specular IOR Level'].default_value = 1
        bsdf.inputs['Specular Tint'].default_value = (1, 1, 1, 1)
        bsdf.inputs['Roughness'].default_value = r
        bsdf.inputs['Anisotropic'].default_value = 0
        bsdf.inputs['Anisotropic Rotation'].default_value = 0.5
        bsdf.inputs['Sheen Weight'].default_value = 0
        bsdf.inputs['Sheen Tint'].default_value = (0, 0, 0, 1)
        bsdf.inputs['Coat Weight'].default_value = 0
        bsdf.inputs['Coat Roughness'].default_value = 0
        bsdf.inputs['Transmission Weight'].default_value = 0
        bsdf.inputs['Emission Color'].default_value = (0, 0, 0, 1)
        bsdf.inputs['Alpha'].default_value = 1
    else:
        bsdf.inputs['Base Color'].default_value = (0, 0, 0, 1)
        bsdf.inputs['Subsurface'].default_value = 0
        bsdf.inputs['Metallic'].default_value = 0
        bsdf.inputs['Specular'].default_value = 1
        bsdf.inputs['Specular Tint'].default_value = 1
        bsdf.inputs['Roughness'].default_value = r
        bsdf.inputs['Anisotropic'].default_value = 0
        bsdf.inputs['Anisotropic Rotation'].default_value = 0.5
        bsdf.inputs['Sheen'].default_value = 0
        bsdf.inputs['Sheen Tint'].default_value = 0
        bsdf.inputs['Clearcoat'].default_value = 0
        bsdf.inputs['Clearcoat Roughness'].default_value = 0
        bsdf.inputs['Transmission'].default_value = 0
        bsdf.inputs['Transmission Roughness'].default_value = 0
        bsdf.inputs['Emission'].default_value = (0, 0, 0, 1)
        bsdf.inputs['Alpha'].default_value = 1

    if override_normal_map:
        override_normal_map_op(material, normal_map_path)

    return material


def create_invisible_material(material_name="Invisible_Material") -> bpy.types.Material:
    """
    Create invisible material

    :param material_name: material name, default is "Invisible_Material"
    :return: invisible material
    """

    material = bpy.data.materials.new(name=material_name)
    material.use_nodes = True
    bsdf = material.node_tree.nodes["Principled BSDF"]

    bsdf.inputs['Alpha'].default_value = 0.

    return material


def create_specular_roughness_bsdf():
    """
    Create specular-roughness workflow BSDF Node Group
    """

    if 'SpecularRoughnessBSDF' in bpy.data.node_groups:
        print('Already created specular roughness workflow BSDF')
        return

    # Create a new node group
    group = bpy.data.node_groups.new(type="ShaderNodeTree", name="SpecularRoughnessBSDF")
    
    # Create group inputs
    group_inputs = group.nodes.new('NodeGroupInput')
    group_inputs.location = (-600, -300)
    
    # Create group outputs
    group_outputs = group.nodes.new('NodeGroupOutput')
    group_outputs.location = (400, 0)
    
    # Create the group input sockets
    roughness_socket = group.interface.new_socket(
        name="Roughness",
        description="Roughness float value",
        in_out='INPUT',
        socket_type='NodeSocketFloat'
    )
    group.interface.items_tree["Roughness"].min_value = 0
    group.interface.items_tree["Roughness"].max_value = 1
    normal_socket = group.interface.new_socket(
        name="Normal",
        description="Normal vector",
        in_out='INPUT',
        socket_type='NodeSocketVector'
    )
    specular_socket = group.interface.new_socket(
        name="Specular",
        description="Specular vector",
        in_out='INPUT',
        socket_type='NodeSocketVector'
    )
    diffuse_socket = group.interface.new_socket(
        name="Diffuse",
        description="Diffuse vector",
        in_out='INPUT',
        socket_type='NodeSocketVector'
    )
    
    output_socket = group.interface.new_socket(
        name="Shader",
        description="Output shader",
        in_out='OUTPUT',
        socket_type='NodeSocketShader'
    )
    
    # Create nodes
    nodes = group.nodes
    links = group.links
    
    # Create Glossy BSDF node
    glossy = nodes.new('ShaderNodeBsdfGlossy')
    glossy.location = (-200, 0)
    glossy.distribution = 'MULTI_GGX'  # Set to Multiscatter GGX
    
    # Create Diffuse BSDF node
    diffuse = nodes.new('ShaderNodeBsdfDiffuse')
    diffuse.location = (-200, -200)
    
    # Create Add Shader node
    add_shader = nodes.new('ShaderNodeAddShader')
    add_shader.location = (200, 0)
    
    # Create links
    
    # Diffuse BSDF connections
    links.new(group_inputs.outputs['Diffuse'], diffuse.inputs['Color'])
    links.new(group_inputs.outputs['Normal'], diffuse.inputs['Normal'])
    diffuse.inputs['Roughness'].default_value = 0.0  # Disable Oren-Nayar BSDF
    
    # Glossy BSDF connections
    links.new(group_inputs.outputs['Specular'], glossy.inputs['Color'])
    links.new(group_inputs.outputs['Roughness'], glossy.inputs['Roughness'])
    links.new(group_inputs.outputs['Normal'], glossy.inputs['Normal'])
    
    # Add Shader connections
    links.new(glossy.outputs[0], add_shader.inputs[0])
    links.new(diffuse.outputs[0], add_shader.inputs[1])
    
    # Final output connection
    links.new(add_shader.outputs[0], group_outputs.inputs['Shader'])
    
    return group


def create_specular_roughness_material(diffuse_color=(1.0, 1.0, 1.0), specular_color=(0.0, 0.0, 0.0), roughness=1.0, material_name="Specular_Roughness_Material") -> bpy.types.Material:
    """
    Create specular-roughness material

    :param diffuse_color: diffuse color, default is (1.0, 1.0, 1.0)
    :param specular_color: specular color, default is (0.0, 0.0, 0.0)
    :param roughness: roughness, default is 1.0
    :param material_name: material name, default is "Specular_Roughness_Material"
    :return: specular-roughness material
    """

    assert IS_BLENDER_4, "This function is only tested in Blender 4.0"
    
    # Create new material
    material = bpy.data.materials.new(name=material_name)
    material.use_nodes = True
    nodes = material.node_tree.nodes
    links = material.node_tree.links
    
    # Clear default nodes
    nodes.clear()
    
    # Create nodes
    specular_rough = nodes.new('ShaderNodeGroup')
    material_output = nodes.new('ShaderNodeOutputMaterial')
    
    # Set up Specular Roughness node
    if 'SpecularRoughnessBSDF' not in bpy.data.node_groups:
        create_specular_roughness_bsdf()
    specular_rough.node_tree = bpy.data.node_groups['SpecularRoughnessBSDF']
    
    # Set diffuse and specular colors
    specular_rough.inputs['Diffuse'].default_value = diffuse_color
    specular_rough.inputs['Specular'].default_value = specular_color
    specular_rough.inputs['Roughness'].default_value = roughness
    
    # Position nodes
    specular_rough.location = (0, 0)
    material_output.location = (300, 0)
    
    # Create links
    links.new(specular_rough.outputs['Shader'], material_output.inputs['Surface'])
    
    return material
